"""
A module for soil moisture anomalies calculation methods based on climatology. The module implements the following methods:

1. Z-Score: The standarized z-score method. 
2. SMAPI: The Soil Moisture Anomaly Percent Index method.
3. SMDI: The Soil Moisture Deficit Index method.
4. ESSMI: The Empirical Standardized Soil Moisture Index method.
5. SMAD: The Standardized Median Absolute Deviation method.
6. SMDS: The Soil Moisture Drought Severity method.
7. SMCI: The Soil Moisture Condition Index method.
8. SMCA: The Soil Moisture Content Anomaly method.
9. ParaDis: The Parametric Distribution method.


"""

__author__ = "Muhammed Abdelaal"
__email__ = "muhammedaabdelaal@gmail.com"

import warnings
from typing import List
import pandas as pd


from smadi.climatology import Climatology
from smadi.preprocess import filter_df, clim_groupping
from smadi.indicators import (
    zscore,
    smapi,
    smdi,
    smad,
    smca,
    smci,
    smds,
    essmi,
    smd,
    para_dis,
)


# Disable RuntimeWarning
warnings.filterwarnings("ignore", category=RuntimeWarning)


class AnomalyDetector(Climatology):
    """
    A base class for detecting anomalies in time series data based on climatology.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "month",
        normal_metrics: List[str] = ["mean"],
    ):
        """
        Initialize the AnomalyDetector class.

        parameters:
        -----------
        df: pd.DataFrame
            A dataframe containing the time series data.

        variable: str
            The name of the variable in the time series data to be analyzed.

        fillna: bool
            A boolean value to indicate whether to fill the missing values in the time series data.

        fillna_window_size: int
            The window size to be used for filling the missing values in the time series data.

        smoothing: bool
            A boolean value to indicate whether to smooth the time series data.

        smooth_window_size: int
            The window size to be used for smoothing the time series data.

        timespan: list[str, str] optional
            The start and end dates for a timespan to be aggregated. Format: ['YYYY-MM-DD ]

        metrics: List[str]
            A list of metrics to be used in the climate normal(climatology) computation. It can be any of the following:
            ['mean', 'median', 'min', 'max']

        groupby_param: List[str]
            The column name to be used for grouping the data for the computation of the climate normal.

        """
        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
            time_step,
            normal_metrics,
        )
        # self.clim_df = pd.DataFrame()
        # self.groupby_param = None

    @property
    def groupby_param(self):

        return clim_groupping(self.clim_df, self.time_step)

    def _preprocess(self, **kwargs) -> pd.DataFrame:
        """
        Preprocess the data before computing the anomalies.
        """
        self._validate_input()
        self.clim_df = self.compute_normals(**kwargs)
        return self.clim_df

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        """
        Base method for detecting anomalies in time series data based on climatology.

        parameters:
        -----------
        kwargs: str
            Date/time parameters to be used for filtering the data before computing the anomalies. It can be any of the following:
            ['year', 'month', 'week', 'dekad' , start_date, end_date]

        returns:
        --------

        pd.DataFrame
            A dataframe containing the computed anomalies.

        """
        self.clim_df = self._preprocess().copy()

        return filter_df(self.clim_df, **kwargs)


class AbsoluteAnomaly(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the absolute deviation value from the climate normal.

    SMA = x - ref

        where:
        x: the average value of the variable in the time series data. It can be any of the following:
        Daily average, weekly average, monthly average, etc.
        ref: the long-term mean (μ) or median (η) of the variable(the climate normal).

    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "month",
        normal_metrics: List[str] = ["mean"],
    ):
        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
            time_step,
            normal_metrics,
        )

    def _preprocess(self, **kwargs) -> pd.DataFrame:
        super()._preprocess(**kwargs)
        return self.clim_df

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        super().detect_anomaly(**kwargs)

        for metric in self.normal_metrics:
            self.clim_df[f"abs-{metric}"] = (
                self.clim_df[f"{self.var}-avg"] - self.clim_df[f"norm-{metric}"]
            )

        return filter_df(self.clim_df, **kwargs)


class ZScore(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Z-Score method.

    z_score = (x - μ) / σ

        where:
        x: the average value of the variable in the time series data. It can be any of the following:
        Daily average, weekly average, monthly average, etc.
        μ: the long-term mean of the variable(the climate normal).
        σ: the long-term standard deviation of the variable.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "month",
    ):
        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
            time_step,
        )

    def _preprocess(self, **kwargs) -> pd.DataFrame:
        super()._preprocess(**kwargs)
        return self.clim_df

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        super().detect_anomaly(**kwargs)

        self.clim_df["zscore"] = self.clim_df.groupby(self.groupby_param)[
            f"{self.var}-avg"
        ].transform(zscore)

        return filter_df(self.clim_df, **kwargs)


class SMAPI(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Soil Moisture Anomaly Percent Index(SMAPI) method.

    SMAPI = ((x - ref) / ref) * 100

    where:
    x: the average value of the variable in the time series data. It can be any of the following:
    Daily average, weekly average, monthly average, etc.
    ref: the long-term mean (μ​) or median (η) of the variable(the climate normal).

    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "month",
        normal_metrics: str = ["mean"],
    ):
        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
            time_step,
            normal_metrics,
        )

    def _preprocess(self, **kwargs) -> pd.DataFrame:
        super()._preprocess(**kwargs)
        return self.clim_df

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        super().detect_anomaly(**kwargs)

        for metric in self.normal_metrics:
            self.clim_df[f"smapi-{metric}"] = self.clim_df.groupby(self.groupby_param)[
                f"{self.var}-avg"
            ].transform(smapi, metric=metric)

        return filter_df(self.clim_df, **kwargs)


class SMDI(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Soil Moisture Deficit Index(SMDI) method.

    SMDI = 0.5 * SMDI(t-1) + (SD(t) / 50)

    where

    SD(t) = ((x - η) / (η - min)) * 100 if x <= η
    SD(t) = ((x - η) / (max - η)) * 100 if x > η

    x: the average value of the variable in the time series data. It can be any of the following:
    Daily average, weekly average, monthly average, etc.
    η: the long-term median of the variable(the climate normal).
    min: the long-term minimum of the variable.
    max: the long-term maximum of the variable.
    t: the time step of the time series data.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "week",
    ):
        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
            time_step,
        )

    def _preprocess(self, **kwargs) -> pd.DataFrame:
        super()._preprocess(**kwargs)
        return self.clim_df

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        super().detect_anomaly(**kwargs)

        self.clim_df["sd"] = self.clim_df.groupby(self.groupby_param)[
            f"{self.var}-avg"
        ].transform(smd)
        self.clim_df["smdi"] = smdi(self.clim_df["sd"])

        return filter_df(self.clim_df, **kwargs)


class SMCA(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Soil Moisture Content Anomaly(SMCA) method.

    SMCA = (x - ref) / (max - ref)

    where:
    x: the average value of the variable in the time series data. It can be any of the following:
    Daily average, weekly average, monthly average, etc.

    ref: the long-term mean (μ) or median (η) of the variable(the climate normal).
    max: the long-term maximum of the variable.
    min: the long-term minimum of the variable.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "month",
        normal_metrics: List[str] = ["mean"],
    ):

        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
            time_step,
            normal_metrics,
        )

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        super().detect_anomaly(**kwargs)
        for metric in self.normal_metrics:
            self.clim_df[f"smca-{metric}"] = self.clim_df.groupby(self.groupby_param)[
                f"{self.var}-avg"
            ].transform(smca, metric=metric)

        return filter_df(self.clim_df, **kwargs)


class SMAD(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Standardized Median Absolute Deviation(SMAD) method.

    SMAD = (x - η) / IQR

    where:
    x: the average value of the variable in the time series data. It can be any of the following:
    Daily average, weekly average, monthly average, etc.
    η: the long-term median of the variable(the climate normal).
    IQR: the interquartile range of the variable. It is the difference between the 75th and 25th percentiles of the variable.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "month",
    ):
        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
            time_step,
            ["median"],
        )

    def _preprocess(self, **kwargs) -> pd.DataFrame:
        super()._preprocess(**kwargs)
        return self.clim_df

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        super().detect_anomaly(**kwargs)
        self.clim_df["smad"] = self.clim_df.groupby(self.groupby_param)[
            f"{self.var}-avg"
        ].transform(smad)

        return filter_df(self.clim_df, **kwargs)


class SMCI(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Soil Moisture Condition Index(SMCI) method.

    SMCI = ((x - min) / (max - min))

    where:
    x: the average value of the variable in the time series data. It can be any of the following:
    Daily average, weekly average, monthly average, etc.
    min: the long-term minimum of the variable.
    max: the long-term maximum of the variable.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "month",
    ):

        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
            time_step,
            ["min", "max"],
        )

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        super().detect_anomaly(**kwargs)

        self.clim_df["smci"] = self.clim_df.groupby(self.groupby_param)[
            f"{self.var}-avg"
        ].transform(smci)

        return filter_df(self.clim_df, **kwargs)


class SMDS(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Soil Moisture Drought Severity(SMDS) method.

    SMDS = 1 - SMP
    SMP = (rank(x) / (n+1))

    where:

    SMP: the Soil Moisture Percentile. It is the percentile of the average value of the variable in the time series data.
    SMDS: the Soil Moisture Drought Severity. It is the severity of the drought based on the percentile of the average value of the variable in the time series data.
    rank(x): the rank of the average value of the variable in the time series data.
    n: the number of years in the time series data.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "month",
    ):
        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
            time_step,
        )

    def _preprocess(self, **kwargs) -> pd.DataFrame:
        super()._preprocess(**kwargs)
        return self.clim_df

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        super().detect_anomaly(**kwargs)
        self.clim_df["smds"] = self.clim_df.groupby(self.groupby_param)[
            f"{self.var}-avg"
        ].transform(smds)
        return filter_df(self.clim_df, **kwargs)


class ESSMI(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on the Empirical Standardized Soil
    Moisture Index(ESSMI) method.

    The index is computed by fitting the nonparametric empirical probability
    density function (ePDF) using the kernel density estimator KDE

    f^h = 1/nh * Σ K((x - xi) / h)
    K = 1/√(2π) * exp(-x^2/2)

    where:
    f^h: the ePDF
    K: the Guassian kernel function
    h: the bandwidth of the kernel function as smoothing parameter (Scott's rule)
    n: the number of observations
    x: the average value of the variable in the time series data. It can be any of the following:
    Daily average, weekly average, monthly average, etc.
    xi: the ith observation

    The ESSMI is then computed by transforming the ePDF to the standard normal distribution with a mean of zero and
    a standard deviation of one using the inverse of the standard normal distribution function.

    ESSMI = Φ^-1(F^h(x))

        where:
        Φ^-1: the inverse of the standard normal distribution function
        F^h: the ePDF


    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "month",
    ):
        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
            time_step,
            ["mean"],
        )

    def _preprocess(self, **kwargs) -> pd.DataFrame:
        super()._preprocess(**kwargs)
        return self.clim_df

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        super().detect_anomaly(**kwargs)
        self.clim_df["essmi"] = self.clim_df.groupby(self.groupby_param)[
            f"{self.var}-avg"
        ].transform(essmi)

        return filter_df(self.clim_df, **kwargs)


class ParaDis(AnomalyDetector):
    """
    A class for detecting anomalies in time series data based on fitting the observed data to a parametric distribution(e.g. beta, gamma, etc.).

    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "month",
        dist: List[str] = ["beta"],
    ):

        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
            time_step,
        )
        self.dist = dist

    def _preprocess(self, **kwargs) -> pd.DataFrame:
        super()._preprocess(**kwargs)
        return self.clim_df

    def detect_anomaly(self, **kwargs) -> pd.DataFrame:
        super().detect_anomaly(**kwargs)
        for dist in self.dist:
            self.clim_df[f"{dist}"] = self.clim_df.groupby(self.groupby_param)[
                f"{self.var}-avg"
            ].transform(para_dis, dist=dist)

        return filter_df(self.clim_df, **kwargs)


if __name__ == "__main__":

    from pathlib import Path
    from smadi.data_reader import extract_obs_ts

    ascat_path = Path("/home/m294/VSA/Code/datasets")

    po = 4854801
    sm_ts = extract_obs_ts(po, ascat_path, obs_type="sm", read_bulk=False)
    df = AbsoluteAnomaly(
        sm_ts, "sm", time_step="month", normal_metrics=["mean", "median"]
    ).detect_anomaly()

    print(df)



"""
run_workflow.py - SMADI Workflow Execution

"""

__author__ = "Muhammed Abdelaal"
__email__ = "muhammedaabdelaal@gmail.com"

from argparse import ArgumentParser, Namespace, ArgumentError
from typing import List, Tuple, Union, Dict
from concurrent.futures import ProcessPoolExecutor
from functools import partial
import pandas as pd
import numpy as np


from smadi.metadata import _Detectors
from smadi.data_reader import AscatData
from smadi.utils import (
    create_logger,
    log_exception,
    log_time,
    load_gpis_by_country,
    get_gpis_from_bbox,
)


def setup_argument_parser() -> ArgumentParser:
    """
    Setup argument parser for SMADI workflow execution.
    """
    parser = ArgumentParser(
        description="Run the SMADI workflow for anomaly detection in ASCAT data"
    )

    # Required arguments
    parser.add_argument(
        "data_path", metavar="data_path", type=str, help="Path to the ASCAT data"
    )
    parser.add_argument(
        "aoi",
        metavar="aoi",
        type=str,
        help="Country name or bounding box coordinates\
        in the tuple 'lon_min, lon_max , lat_min, lat_max'",
    )
    parser.add_argument(
        "time_step",
        metavar="time_step",
        type=str,
        default="month",
        choices=["month", "dekad", "week", "bimonth", "day"],
        help="The time step for the climatology calculation. Supported values: month, dekad, week, bimonth, day",
    )

    # Optional arguments
    parser.add_argument(
        "--variable",
        metavar="variable",
        type=str,
        default="sm",
        help="The variable to be used for the anomaly detection.",
    )
    parser.add_argument(
        "--year",
        metavar="year",
        type=int,
        nargs="*",
        default=None,
        required=True,
        choices=range(2007, 2023),
        help="The year(s) for the date parameters",
    )
    parser.add_argument(
        "--month",
        metavar="month",
        type=int,
        nargs="*",
        default=None,
        choices=range(1, 13),
        help="The month(s) for the date parameters",
    )
    parser.add_argument(
        "--dekad",
        metavar="dekad",
        type=int,
        nargs="*",
        default=None,
        choices=(1, 2, 3),
        help="The dekad(s) for the date parameters",
    )
    parser.add_argument(
        "--week",
        metavar="week",
        type=int,
        nargs="*",
        default=None,
        choices=range(1, 53),
        help="The week(s) for the date parameters",
    )
    parser.add_argument(
        "--bimonth",
        metavar="bimonth",
        type=int,
        nargs="*",
        default=None,
        choices=(1, 2),
        help="The bimonth(s) for the date parameters",
    )
    parser.add_argument(
        "--day",
        metavar="day",
        type=int,
        nargs="*",
        default=None,
        choices=range(1, 32),
        help="The day(s) for the date parameters",
    )
    parser.add_argument(
        "--methods",
        metavar="methods",
        type=str,
        nargs="*",
        default=["zscore"],
        help="Anomaly detection methods. Supported methods: zscore, smapi-mean,\
            smapi-median, smdi, smca-mean, smca-median, smad, smci, smds, essmi,\
                beta, gamma, abs-mean, abs-median",
    )
    parser.add_argument(
        "--timespan",
        metavar="timespan",
        type=list,
        default=None,
        help="To work on a subset of the data. Example: ['2012-01-01', '2012-12-31']",
    )
    parser.add_argument(
        "--fillna", type=bool, default=False, help="Fill missing values"
    )
    parser.add_argument("--fillna_size", type=int, default=3, help="Fillna window size")
    parser.add_argument("--smoothing", type=bool, default=False, help="Apply smoothing")
    parser.add_argument(
        "--smooth_size", type=int, default=31, help="Smoothing window size"
    )
    parser.add_argument(
        "--workers",
        metavar="workers",
        type=int,
        default=None,
        help="The number of workers to use for multiprocessing",
    )
    parser.add_argument(
        "--save_to",
        type=str,
        default=None,
        help="Save the output to a file to the given path",
    )

    return parser


parser = setup_argument_parser()
args: Namespace = parser.parse_args()

# Required parameters
data_path = args.data_path
aoi = args.aoi
if "," in aoi:
    aoi = tuple(map(float, aoi.split(",")))
variable = args.variable
time_step = args.time_step

# Optional parameters
methods = args.methods
workers = args.workers
fillna = args.fillna
fillna_window_size = args.fillna_size
smoothing = args.smoothing
smooth_window_size = args.smooth_size
timespan = args.timespan
year = args.year
month = args.month
dekad = args.dekad
week = args.week
bimonth = args.bimonth
day = args.day
save_to = args.save_to

# Create an instance of the AscatData class
ascat_obj = AscatData(data_path, False)

# Create a logger
logger = create_logger("run_logger")


@log_exception(logger)
def load_ts(gpi, variable="sm"):
    """
    Load ASCAT time series for a given gpi
    """
    ascat_ts = ascat_obj.read(gpi)
    valid = ascat_ts["num_sigma"] >= 2
    ascat_ts.loc[~valid, ["sm", "sigma40", "slope40", "curvature40"]] = np.nan
    df = pd.DataFrame(ascat_ts.get(variable))
    return df


def _validate_param(param_name: str, param_value: List[int]) -> None:

    if param_value is None:
        raise ValueError(f"The '{param_name}' parameter must be provided")
    if not (isinstance(param_value, (list, int))):
        raise ValueError(f"The '{param_name}' parameter must be an int of list of ints")


def _validate_required_params(
    time_step: str,
    required_params: Dict[str, List[str]],
    local_vars: Dict[str, List[int]],
) -> None:

    missing_params = [
        param for param in required_params[time_step] if local_vars.get(param) is None
    ]
    if missing_params:
        raise ValueError(
            f"For time_step '{time_step}', the following parameters must be provided: "
            f"{', '.join(missing_params)}"
        )


def validate_date_params(
    time_step: str,
    year: Union[int, List[int]] = None,
    month: Union[int, List[int]] = None,
    dekad: Union[int, List[int]] = None,
    week: Union[int, List[int]] = None,
    bimonth: Union[int, List[int]] = None,
    day: Union[int, List[int]] = None,
) -> Dict[str, List[int]]:
    """
    Validate the date parameters for the anomaly detection workflow.
    """

    year = [year] if isinstance(year, int) else year
    month = [month] if isinstance(month, int) else month
    dekad = [dekad] if isinstance(dekad, int) else dekad
    week = [week] if isinstance(week, int) else week
    bimonth = [bimonth] if isinstance(bimonth, int) else bimonth
    day = [day] if isinstance(day, int) else day

    date_param = {"year": year, "month": month}

    # Validating dekad or week based on time_step

    if time_step == "month":
        pass
    elif time_step == "dekad":
        date_param["dekad"] = dekad
    elif time_step == "bimonth":
        date_param["bimonth"] = bimonth

    elif time_step == "day":
        date_param["day"] = day

    elif time_step == "week":
        date_param["week"] = week
        date_param.pop("month")

    else:
        raise ValueError(
            f"Unsupported time_step: {time_step}. Supported time_steps are month, dekad, week, bimonth, day"
        )

    # Validation for parameters
    for param_name, param_value in date_param.items():
        _validate_param(param_name, param_value)

    # Checking if the value lists are of the same length
    if len(set(map(len, date_param.values()))) > 1:
        raise ValueError(
            "The length of the date parameters lists must be the same for multiple dates"
        )

    # Checking if required parameters are provided
    required_params = {
        "month": ["year", "month"],
        "dekad": ["year", "month", "dekad"],
        "week": ["year", "week"],
        "bimonth": ["year", "month", "bimonth"],
        "day": ["year", "month", "day"],
    }

    local_vars = locals()
    _validate_required_params(time_step, required_params, local_vars)

    return date_param


def validate_anomaly_method(methods):
    """
    Validate the anomaly detection method
    """

    for method in methods:
        if method not in _Detectors.keys():
            raise ValueError(
                f"Anomaly method '{method}' is not supported."
                f"Supported methods are one of the following: {tuple(_Detectors.keys())}"
            )


@log_exception(logger)
def single_po_run(
    gpi: int,
    methods: str = ["zscore"],
    variable: str = "sm",
    time_step: str = "month",
    fillna: bool = False,
    fillna_window_size: int = None,
    smoothing: bool = False,
    smooth_window_size: int = None,
    year: Union[int, List[int]] = None,
    month: Union[int, List[int]] = None,
    dekad: Union[int, List[int]] = None,
    week: Union[int, List[int]] = None,
    bimonth: Union[int, List[int]] = None,
    day: Union[int, List[int]] = None,
    timespan: List[str] = None,
) -> Tuple[int, Dict[str, float]]:
    """
    Run the anomaly detection workflow for a single grid point index.
    """

    # Load the time series for the given gpi
    global ascat_obj
    df = load_ts(gpi, variable=variable)
    # Validate the date parameters
    date_params = validate_date_params(
        time_step, year, month, dekad, week, bimonth, day
    )
    # Create a list of dictionaries containing the date parameters
    date_params = [
        dict(zip(date_params.keys(), values)) for values in zip(*date_params.values())
    ]

    # Define a dictionary to store the results
    results = {}
    for method in methods:

        # Define the anomaly detection parameters
        anomaly_params = {
            "df": df,
            "variable": variable,
            "time_step": time_step,
            "fillna": fillna,
            "fillna_window_size": fillna_window_size,
            "smoothing": smoothing,
            "smooth_window_size": smooth_window_size,
            "timespan": timespan,
        }

        # If the method has a metric parameter (e.g. smapi-mean, smapi-median), set the metric parameter
        if "-" in method:
            anomaly_params["normal_metrics"] = [method.split("-")[1]]

        elif method in ["beta", "gamma"]:
            anomaly_params["dist"] = [method]

        try:
            for date_param in date_params:
                anomaly = _Detectors[method](**anomaly_params).detect_anomaly(
                    **date_param
                )
                date_str = f"-".join(str(value) for value in date_param.values())
                results[method + f"({date_str})"] = anomaly[method].values[0]

        except AttributeError as e:
            return None

    return (gpi, results)


@log_exception(logger)
def _finalize(result: Tuple[int, dict], df: pd.DataFrame, gpis_col="point"):
    try:
        gpi, anomaly = result
    except Exception as e:
        return df

    else:
        for method, value in anomaly.items():
            df.loc[df[gpis_col] == gpi, method] = value

    return df


@log_time(logger)
def run(
    aoi: Union[str, Tuple[float, float, float, float]],
    methods: Union[str, List[str]] = ["zscore"],
    variable: str = "sm",
    time_step: str = "month",
    fillna: bool = False,
    fillna_window_size: int = None,
    smoothing: bool = False,
    smooth_window_size: int = None,
    timespan: List[str] = None,
    year: List[int] = None,
    month: List[int] = None,
    dekad: List[int] = None,
    week: List[int] = None,
    bimonth: List[int] = None,
    day: List[int] = None,
    workers: int = None,
) -> pd.DataFrame:
    """
    Run the anomaly detection workflow for multiple grid point indices with multiprocessing support.
    """
    # Print workflow start message
    print("\nWorkflow started....\n")
    print(f"Loading grid points for {aoi}....")

    if isinstance(aoi, str):
        pointlist = load_gpis_by_country(aoi)
    else:
        pointlist = get_gpis_from_bbox(aoi)

    print(f"Grid points loaded successfully for {aoi}\n")
    print(pointlist.head())
    print("\n")
    pointlist = pointlist[:100]
    pre_compute = partial(
        single_po_run,
        methods=methods,
        variable=variable,
        time_step=time_step,
        fillna=fillna,
        fillna_window_size=fillna_window_size,
        smoothing=smoothing,
        smooth_window_size=smooth_window_size,
        year=year,
        month=month,
        dekad=dekad,
        week=week,
        bimonth=bimonth,
        day=day,
        timespan=timespan,
    )

    # Print the workflow initiation message
    print(f"Running the anomaly detection workflow for {aoi}....\n")

    # Print workflow parameters
    print("Workflow parameters:\n")
    print(f"    Anomaly detection methods: {', '.join(methods)}")
    print(f"    Variable: {variable}")
    print(f"    Time step for Climatology: {time_step}")
    print(f"    Date parameters:\n")

    # Print date parameters if available
    date_parameters = {
        "Year": year,
        "Month": month,
        "Dekad": dekad,
        "Week": week,
        "Bimonth": bimonth,
        "Day": day,
    }
    for param, value in date_parameters.items():
        if value:
            print(f"     {param}: {value}")

    with ProcessPoolExecutor(workers) as executor:
        results = executor.map(pre_compute, pointlist["point"])
        for result in results:
            pointlist = _finalize(result, pointlist)

        return pointlist


def main():

    try:
        validate_anomaly_method(methods)
        validate_date_params(time_step, year, month, dekad, week, bimonth, day)
    except ArgumentError as e:
        parser.error(str(e))

    from time import time

    start = time()
    df = run(
        aoi=aoi,
        methods=methods,
        variable=variable,
        time_step=time_step,
        year=year,
        month=month,
        week=week,
        day=day,
        bimonth=bimonth,
        dekad=dekad,
        timespan=timespan,
        fillna=fillna,
        fillna_window_size=fillna_window_size,
        smoothing=smoothing,
        smooth_window_size=smooth_window_size,
        workers=workers,
    )

    print(f"\nWorkflow completed in {time() - start} seconds\n")
    print(df)

    if save_to:
        try:
            df.to_csv(save_to)
            print(f"Saving the output data frame to {save_to}....")
            print("Output saved successfully")

        except ArgumentError as e:
            parser.error(str(e))


if __name__ == "__main__":
    main()



"""
A module for calculating climatology (climate normal) for different time steps (month, dekad, week) based on time series data.
"""

__author__ = "Muhammed Abdelaal"
__email__ = "muhammedaabdelaal@gmail.com"

from typing import List
import pandas as pd


from smadi.preprocess import (
    fillna,
    smooth,
    filter_df,
    monthly_agg,
    dekadal_agg,
    weekly_agg,
    bimonthly_agg,
    compute_clim,
)


class Aggregator:
    """
    Base class for aggregation

    Attributes:
    -----------
    df : pd.DataFrame
        The DataFrame containing the data to be aggregated.

    variable : str
        The variable/column in the DataFrame to be aggregated.

    fillna : bool
        Fill NaN values in the time series data using a moving window average.

    fillna_window_size : int
        The size of the moving window for filling NaN values. It is recommended to be an odd number.

    smoothing : bool
        Smooth the time series data using a moving window average.

    smooth_window_size : int
        The size of the moving window for smoothing(n-days). It is recommended to be an odd number.

    timespan : list[str, str] optional
        The start and end dates for a timespan to be aggregated. Format: ['YYYY-MM-DD', 'YYYY-MM-DD']

    resulted_df : pd.DataFrame
        The resulting DataFrame after aggregation.

    Methods:
    --------

    _fillna:
        Fills NaN values in the time series data using a moving window average.

    _smooth:
        Smooths the time series data using a moving window average.

    _set_up_mode():
        Filters the DataFrame based on the parameters provided to perform aggregation on a subset or all of the data.

    _filter_df:
        Filters the DataFrame based on specified time/date conditions.

    _validate_df_index:
        Validates the input DataFrame type and index.

    _validate_variable:
        Validates the variable to be aggregated.

    _validate_input:
        Validates the input parameters.

    aggregate:
        Aggregates the data based on the specified time step.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
    ):
        """
        Initializes the Aggregation class.

        """

        self.original_df = df
        self.var = variable
        self.fillna = fillna
        self.fillna_window_size = fillna_window_size
        self.smoothing = smoothing
        self.smooth_window_size = smooth_window_size
        self.timespan = timespan

        self._validate_input()
        self.df = pd.DataFrame(self.original_df[self.var]).resample("D").mean()
        self.df = (
            self.df.truncate(before=timespan[0], after=timespan[1])
            if timespan
            else self.df
        )
        self._fillna()
        self._smooth()
        self.df.dropna(inplace=True)
        self.resulted_df = pd.DataFrame()

    def _fillna(self):
        """
        Fills NaN values in the time series data using a moving window average.
        """
        if self.fillna:
            self.df[self.var] = fillna(self.df, self.var, self.fillna_window_size)

    def _smooth(self):
        """
        Smooths the time series data using a moving window average.
        """
        if self.smoothing:
            self.df[self.var] = smooth(self.df, self.var, self.smooth_window_size)

    def _validate_df_index(self):
        """
        Validates the input DataFrame type and index.

        Raises:
        -------

        TypeError:
            If the input DataFrame is not a pandas DataFrame.

        ValueError:
            If the input DataFrame index is not a datetime index.
        """
        if not isinstance(self.original_df, pd.DataFrame):
            raise TypeError("df must be a pandas DataFrame")

        if not isinstance(self.original_df.index, pd.DatetimeIndex):
            raise ValueError("df index must be a datetime index")

    def _validate_variable(self):
        """
        Validates the variable to be aggregated.

        Raises:
        -------
        ValueError:
            If the variable is not found in the input DataFrame columns.

        """
        if self.var not in self.original_df.columns:
            raise ValueError(
                f"Variable '{self.var}' not found in the input DataFrame columns."
            )

    def _validate_fillna_smoothing(self):
        """
        Validates the smoothing parameters.

        Raises:
        -------
        ValueError:
            - If the window size is not provided when smoothing is enabled.
        TypeError:
            - If the smoothing parameter is not a boolean value.
            - If the window size parameter is not an integer value when smoothing is enabled.
        """

        if any(
            [
                self.fillna and self.fillna_window_size is None,
                self.smoothing and self.smooth_window_size is None,
            ]
        ):

            raise ValueError(
                "window size must be provided when 'fillna' or 'smoothing' is enabled"
            )

    def _validate_input(self):
        """
        Validates the input parameters.
        """
        self._validate_df_index()
        self._validate_variable()
        self._validate_fillna_smoothing()

    def aggregate(self, **kwargs):
        """
        Aggregates the data based on the specified .
        """
        return filter_df(self.df, **kwargs)


class MonthlyAggregator(Aggregator):
    """
    Aggregates the time series data based on month-based time step.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = 3,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
    ):
        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
        )

    def aggregate(self, **kwargs):

        self.resulted_df[f"{self.var}-avg"] = monthly_agg(self.df, self.var)

        return filter_df(self.resulted_df, **kwargs)


class DekadalAggregator(Aggregator):
    """
    Aggregates the data based on dekad-based time step.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
    ):

        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
        )

    def aggregate(self, **kwargs):

        self.resulted_df[f"{self.var}-avg"] = dekadal_agg(self.df, self.var)

        return filter_df(self.resulted_df, **kwargs)


class WeeklyAggregator(Aggregator):
    """
    Aggregates the time series data based on week-based time step.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
    ):
        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
        )

    def aggregate(self, **kwargs):

        self.resulted_df[f"{self.var}-avg"] = weekly_agg(self.df, self.var)

        return filter_df(self.resulted_df, **kwargs)


class BimonthlyAggregator(Aggregator):
    """
    Aggregates the time series data based on bimonthly (twice a month) time step.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
    ):
        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
        )

    def aggregate(self, **kwargs):

        self.resulted_df[f"{self.var}-avg"] = bimonthly_agg(self.df, self.var)

        return filter_df(self.resulted_df, **kwargs)


class DailyAggregator(Aggregator):
    """
    Aggregates the time series data based on daily time step.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
    ):

        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
        )

    def aggregate(self, **kwargs):
        self.resulted_df[f"{self.var}-avg"] = self.df[self.var]
        return filter_df(self.resulted_df.drop_duplicates(), **kwargs)


class Climatology(Aggregator):
    """
    A class for calculating climatology(climate normal) for time series data.

    Attributes:
    ----------
    df_original: pd.DataFrame
        The original input DataFrame before resampling and removing NaN values.

    df: pd.DataFrame
        The input DataFrame containing the preprocessed data to be aggregated.

    variable: str
        The variable/column in the DataFrame to be aggregated.

    fillna: bool
        Fill NaN values in the time series data using a moving window average.

    fillna_window_size: int
        The size of the moving window for filling NaN values. It is recommended to be an odd number.

    smoothing: bool
        Smooth the time series data using a moving window average.

    smooth_window_size: int
        The size of the moving window for smoothing(n-days). It is recommended to be an odd number.

    timespan: list[str, str] optional
        The start and end dates for a timespan to be aggregated. Format: ['YYYY-MM-DD ]

    time_step: str
        The time step for aggregation. Supported values: 'day', 'week', 'dekad', 'bimonth', 'month'.

    normal_metrics: List[str]
        The metrics to be used in the climatology computation. Supported values: 'mean', 'median', 'min', 'max', etc.

    clima_df: pd.DataFrame
        The DataFrame containing climatology information.

    Methods:
    -------
    aggregate:
        Aggregates the data based on the time step and metrics provided.

    _validate_time_step:
        Validates the time step.

    _validate_metrics:
        Validates the metrics to be used in the climatology computation.

    compute_normals:
        Calculates climatology based on the aggregated data.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        variable: str,
        fillna: bool = False,
        fillna_window_size: int = None,
        smoothing=False,
        smooth_window_size=None,
        timespan: List[str] = None,
        time_step: str = "month",
        normal_metrics: List[str] = ["mean"],
    ):
        """
        Initializes the Climatology class.
        """
        self.time_step = time_step
        self.normal_metrics = normal_metrics
        self.valid_time_steps = ["month", "dekad", "week", "day", "bimonth"]
        self.valid_metrics = ["mean", "median", "min", "max"]
        super().__init__(
            df,
            variable,
            fillna,
            fillna_window_size,
            smoothing,
            smooth_window_size,
            timespan,
        )
        self.clim_df = pd.DataFrame()

    def _validate_time_step(
        self,
    ) -> None:
        """
        Validates the time step.

        Raises:
        -------
        ValueError:
            If the time step is not one of the supported values.

        """
        if self.time_step not in self.valid_time_steps:
            raise ValueError(
                f"Invalid time step '{self.time_step}'. Supported values: {self.valid_time_steps}."
            )

    def _validate_metrics(self):
        """
        Validates the metrics to be used in the climatology computation.

        Raises:
            ValueError: If the metric is not one of the supported values.

        """
        for metric in self.normal_metrics:
            if metric not in self.valid_metrics:
                raise ValueError(
                    f"Invalid metric '{metric}'. Supported values: {self.valid_metrics}."
                )

    def _validate_input(self):
        super()._validate_input()
        self._validate_time_step()
        self._validate_metrics()

    def aggregate(self):
        """
        Aggregates the data based on the time step and metrics provided.

        """
        if self.time_step == "month":
            return MonthlyAggregator(
                self.df,
                self.var,
                self.fillna,
                self.fillna_window_size,
                self.smoothing,
                self.smooth_window_size,
                self.timespan,
            ).aggregate()

        elif self.time_step == "week":
            return WeeklyAggregator(
                self.df,
                self.var,
                self.fillna,
                self.fillna_window_size,
                self.smoothing,
                self.smooth_window_size,
                self.timespan,
            ).aggregate()

        elif self.time_step == "dekad":
            return DekadalAggregator(
                self.df,
                self.var,
                self.fillna,
                self.fillna_window_size,
                self.smoothing,
                self.smooth_window_size,
                self.timespan,
            ).aggregate()

        elif self.time_step == "bimonth":
            return BimonthlyAggregator(
                self.df,
                self.var,
                self.fillna,
                self.fillna_window_size,
                self.smoothing,
                self.smooth_window_size,
                self.timespan,
            ).aggregate()

        elif self.time_step == "day":
            return DailyAggregator(
                self.df,
                self.var,
                self.fillna,
                self.fillna_window_size,
                self.smoothing,
                self.smooth_window_size,
                self.timespan,
            ).aggregate()

    def compute_normals(self, **kwargs) -> pd.DataFrame:
        """
        Calculates climatology based on the aggregated data.

        Parameters:
        -----------
        kwargs:
            Additional time/date filtering parameters.

        Returns:
        --------
        pd.DataFrame
            The DataFrame containing climatology information.
        """
        self.clim_df = compute_clim(
            self.aggregate(), self.time_step, f"{self.var}-avg", self.normal_metrics
        )

        return filter_df(self.clim_df, **kwargs)



import logging
import requests
from functools import wraps
from time import time
from io import StringIO
import pycountry
import pandas as pd

from fibgrid.realization import FibLandGrid


def create_logger(name, level=logging.DEBUG):
    """
    Create a logger with the given name and level

    parameters:
    ----------

    name: str
        name of the logger
    level: logging.LEVEL
        level of the logger

    returns:
    -------
    logger: logging.logger
        a logger object
    """

    # create logger
    logger = logging.getLogger(name)
    logger.setLevel(level)

    # create a file to store all the logs exceptions
    logfile = logging.FileHandler("run_logger.log")

    # create a formatter and set the formatter for the handler
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    logfile.setFormatter(formatter)
    logger.addHandler(logfile)

    return logger


def log_exception(logger):
    """
    A decorator to log exceptions in a function

    parameters:
    ----------
    logger: logging.logger
        a logger object

    returns:
    -------
    decorator: function
        a decorator function

    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                issue = f"Exception '{e}'in " + func.__name__ + "\n"
                issue += "=============================\n"
                logger.exception(issue)

        return wrapper

    return decorator


def log_time(logger):
    """
    A decorator to log the time taken by a function

    parameters:
    ----------

    logger: logging.logger
        a logger object

    returns:
    -------

    decorator: function
        a decorator function

    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start = time()
            result = func(*args, **kwargs)
            end = time()
            logger.info("Function %s took %s seconds", func.__name__, end - start)
            return result

        return wrapper

    return decorator


def get_country_code(country_name):
    """
    Get the ISO 3166-1 alpha-3 country code for a given country name.

    parameters:
    ----------

    country_name: str
        name of the country

    returns:
    -------

    country_code: str
        ISO 3166-1 alpha-3 country code
    """
    try:
        # Get ISO alpha-3 code for the provided country name
        country = pycountry.countries.lookup(country_name)
        return country.alpha_3
    except LookupError:
        print("Country name not found.")
        return None


def load_gpis_by_country(country, res=6.25, format="csv"):
    """
    Load the GPIS based on the country name from the DGG API
    Source: https://dgg.geo.tuwien.ac.at/

    parameters:
    ----------
    country: str
        name of the country

    grid: str
        name of the grid to be used. Default is "fibgrid_n6600000". Supported grids are:
        - fibgrid_n6600000 (Fibonacci 6.5 km)
        - fibgrid_n1650000 (Fibonacci 12.5 km)
        - fibgrid_n430000  (Fibonacci 25 km)
        - warp (WARP)

    format: str
        format of the data to be returned. Default is "csv". Supported formats are:
        - csv
        - json
    """

    country = get_country_code(country)

    if res == 6.25:
        grid = "fibgrid_n6600000"
    elif res == 12.5:
        grid = "fibgrid_n1650000"
    elif res == 25:
        grid = "fibgrid_n430000"

    # Construct the URL based on the provided country name, grid, and format
    url = f"https://dgg.geo.tuwien.ac.at/get_points/?grid={grid}&country={country.upper()}&format={format}"

    try:
        response = requests.get(url)

        if response.status_code == 200:
            csv_data = response.content.decode("utf-8")
            df = pd.read_csv(StringIO(csv_data))
            return df
        else:
            print(f"Failed to download CSV file. Status code: {response.status_code}")
            return None
    except Exception as e:
        print(f"An error occurred: {str(e)}")
        return None


def get_gpis_from_bbox(bbox, res=6.25):

    lonmin, lonmax, latmin, latmax = bbox
    grid = FibLandGrid(res)
    gpis = grid.get_bbox_grid_points(
        lonmin=lonmin, lonmax=lonmax, latmin=latmin, latmax=latmax, both=True
    )

    return pd.DataFrame({"point": gpis[0], "lon": gpis[1], "lat": gpis[2]})



import matplotlib
import eomaps

from smadi.metadata import indicators_thresholds

matplotlib.use("Qt5Agg")


def set_thresholds(method):
    """
    Set the thresholds for the specified method based on the method name.

    parameters:
    -----------

    method: str
        The method name for which the thresholds are to be set. Supported methods are:
        'zscore', 'smapi', 'smdi', 'smca', 'smad', 'smci', 'smds', 'essmi', 'beta', 'gamma'
    """

    if method in ["beta", "gamma", "essmi", "zscore", "smad"]:
        return indicators_thresholds["zscore"]
    else:
        return indicators_thresholds[method]


def set_extent(df, x="lon", y="lat", buffer=2):
    """
    Set the extent for the map based on the provided dataframe and buffer.

    parameters:
    -----------

    df: pd.DataFrame
        The dataframe containing the data.

    x: str
        The column name for the x-axis.

    y: str
        The column name for the y-axis.

    buffer: int
        The buffer to be added to the min and max values of the x and y axis.
    """

    min_lat = df[y].min() - buffer
    max_lat = df[y].max() + buffer
    min_lon = df[x].min() - buffer
    max_lon = df[x].max() + buffer

    return min_lon, max_lon, min_lat, max_lat


def set_bins(colm):
    """
    Set the bins and labels for color classification for the selected column.

    parameters:
    -----------

    colm: str
        The data column name for which the bins and labels are to be set.
    """
    method = colm.split("(")[0]
    if "-" in method:
        method = method.split("-")[0]

    thrsholds = set_thresholds(method)
    bins = [val[1] for val in thrsholds.values()]
    labels = [key for key in thrsholds.keys()]
    labels.insert(0, labels[0])
    bins.insert(0, next(iter(thrsholds.values()))[0])

    return bins, labels


def plot_anomaly_map(
    df,
    colm,
    x="lon",
    y="lat",
    crs=4326,
    set_extent_to=True,
    extent=None,
    figsize=(7, 7),
    title="SM Anomaly",
    bins=None,
    labels=None,
    add_gridlines=True,
    g_kwargs={
        "d": (2, 2),
        "ec": "grey",
        "ls": "--",
        "lw": 0.01,
    },
    vmin=None,
    vmax=None,
    add_colorbar=True,
    cmap="RdYlBu",
    cb_kwargs={
        "pos": 0.4,
        "labelsize": 7,
        "tick_lines": "center",
        "show_values": False,
    },
):
    """
    Plot the anomaly map for the selected column from the dataframe.

    parameters:
    -----------

    df: pd.DataFrame
        The dataframe containing the data.

    colm: str
        The column name for the data to be plotted.

    x: str
        The column name for the x-axis.

    y: str
        The column name for the y-axis.

    crs: int
        The coordinate reference system for the map.

    set_extent_to: bool
        Whether to set the map to a specific extent.

    extent: tuple
        The extent to set the map to. if None, the extent is calculated based on the data.

    figsize: tuple
        The figure size for the map.

    title: str
        The title for the map.

    bins: list
        The bins for the color classification.

    labels: list
        The labels for the bins of the color classification.

    add_gridlines: bool
        Whether to add gridlines to the map.

    g_kwargs: dict
        The gridlines keyword arguments. This includes the linestyle, linewidth, edgecolor, and distance between gridlines.

    vmin: float
        The minimum value for the color classification.

    vmax: float
        The maximum value for the color classification.

    cmap: str
        The colormap for the map.

    add_colorbar: bool
        Whether to add a colorbar to the map.

    cb_kwargs: dict
        The colorbar keyword arguments. This includes the position, label size, tick lines, and whether to show values.


    """

    # Initialize map object
    m = eomaps.Maps(figsize=figsize, frameon=True, crs=crs)
    m.ax.set_title(title, pad=20, linespacing=1.5)
    m.set_shape.rectangles(radius=0.05)
    m.set_frame(linewidth=0.5)

    # Set data and parameter
    m.set_data(data=df, parameter=colm, x=x, y=y, crs=crs)

    # Set extent
    if set_extent_to:
        extent = set_extent(df, x=x, y=y) if extent is None else extent
        m.ax.set_extent(extent, eomaps.Maps.CRS.PlateCarree())

    # Set bins and labels
    if bins is None and labels is None:
        bins, labels = set_bins(colm)

    # Set classification
    m.set_classify.UserDefined(bins=bins)
    # Add gridlines
    if add_gridlines:
        g = m.add_gridlines(**g_kwargs)
        g.add_labels(fontsize=8)

    # Set vmin and vmax based on the thresholds min and max
    vmin = bins[0] if vmin is None else vmin
    vmax = bins[-1] if vmax is None else vmax

    # Plot map
    cmap = cmap + "_r" if colm.split("(")[0] == "smds" else cmap
    m.plot_map(vmin=vmin, vmax=vmax, cmap=cmap, lw=1.5)

    # Add colorbar

    if add_colorbar:
        cb = m.add_colorbar(
            pos=cb_kwargs["pos"],
        )
        cb.tick_params(labelsize=cb_kwargs["labelsize"])
        cb.set_bin_labels(
            bins=bins,
            names=labels,
            tick_lines=cb_kwargs["tick_lines"],
            show_values=cb_kwargs["show_values"],
        )

    m.show()
